package ochisuke.vmg2eml;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.net.QuotedPrintableCodec;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.DirectoryDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.ProgressBar;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

import swing2swt.layout.BorderLayout;

public class Vmg2eml implements Runnable {

	protected static final String VERSION = "2.1";
	protected static final String PROGRAM = "vmg2eml ver." + VERSION;
	private static final File SETTING_FILE = new File("./vmg2eml.xml");
	private static final String LOG_FILENAME = "vmg2eml.log";
	private static final String exec_str = "実行";
	private static final String stop_str = "中止";
	static final Pattern regex_tel_from_subject = Pattern.compile("(0\\d*)\\D*");
	static final String CHARSET = "MS932";

	private final QuotedPrintableCodec qpCodec = new QuotedPrintableCodec(CHARSET);

	/**
	 * VMGファイルだけを取得するフィルタ
	 */
	private final FilenameFilter vmgFileFilter = new FilenameFilter() {
		@Override
		public boolean accept(File dir, String name) {
			return name.toLowerCase().endsWith(".vmg");
		}
	};

	/**
	 * 別スレッドから実行ボタンに戻すタスク
	 */
	private final Runnable run_button_exec = new Runnable() {
		@Override
		public void run() {
//			if (interrupted.get()) {
//				progressBar.setState(SWT.ERROR);
//				progressBar.redraw();
//			}
			th_exec = null;
			button_exec.setText(exec_str);
			button_exec.setEnabled(true);
//			System.out.println("Finished.");
		}
	};

	private final Vmg2eml myself;
	private Display display;
	protected Shell mainShell;
	Text text_dir;
	Text text_usertel;
	Text text_useraddr;
	Button btnCheckButton;
	private ProgressBar progressBar;
	private Button button_exec;
	private Text text_log;

	// 処理実行時に使う変数群
	File targetDir;
	boolean complement_mode;
	String usertel;
	String useraddr;
	String sentbox;

	// 処理実行スレッド関連
	Thread th_exec;
	private final AtomicBoolean interrupted;
	private final LogPrinter logPrinter; // ログを画面表示するタスク
	private final AtomicInteger progressPercent;

	/**
	 * 設定項目
	 * @author ochi
	 *
	 */
	private enum Parameter {
		DIR("dir", ""),
		COMPLEMENT("complement", "true"),
		USERTEL("usertel", "0x0xxxxxxxx"),
		USERADDR("useraddr", "xxx@docomo.ne.jp"),
		SENTBOX("sentbox", "SENTBOX")
		;

		final String propkey, defvalue;

		Parameter(String propkey, String defvalue) {
			this.propkey = propkey;
			this.defvalue =defvalue;
		}
	}

	public Vmg2eml() {
		myself = this;
		interrupted = new AtomicBoolean(false);
		logPrinter = new LogPrinter();
		progressPercent = new AtomicInteger(0);
	}


	/**
	 * Launch the application.
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			Vmg2eml window = new Vmg2eml();
			window.open();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Open the window.
	 */
	public void open() {
		display = Display.getDefault();
		createContents();
		loadSetting();
		mainShell.open();
		mainShell.layout();
		while (!mainShell.isDisposed()) {
			if (!display.readAndDispatch()) {
				display.sleep();
			}
		}
	}

	/**
	 * Create contents of the window.
	 */
	protected void createContents() {
		mainShell = new Shell();
		mainShell.setSize(450, 450);
		mainShell.setText(PROGRAM);
		mainShell.setLayout(new BorderLayout(0, 0));

		Composite composite_main = new Composite(mainShell, SWT.NONE);
		composite_main.setLayoutData(BorderLayout.NORTH);
		composite_main.setLayout(new BorderLayout(0, 0));

		Composite composite_setting = new Composite(composite_main, SWT.NONE);
		composite_setting.setLayoutData(BorderLayout.CENTER);
		composite_setting.setLayout(new GridLayout(3, false));

		Label lblHennkannSitai = new Label(composite_setting, SWT.NONE);
		lblHennkannSitai.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false, 1, 1));
		lblHennkannSitai.setText("変換元フォルダ");

		text_dir = new Text(composite_setting, SWT.BORDER);
		text_dir.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));

		Button button_input = new Button(composite_setting, SWT.NONE);
		button_input.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				DirectoryDialog dialog = new DirectoryDialog(mainShell);
				dialog.setFilterPath(text_dir.getText());
				dialog.setMessage("変換元のフォルダを指定してください。");
				String dir = dialog.open();
				if (dir == null) return;
				text_dir.setText(dir);
			}
		});
		button_input.setText("...");

		Group group_complement = new Group(composite_setting, SWT.NONE);
		group_complement.setText("補完設定");
		group_complement.setLayout(new GridLayout(2, false));
		group_complement.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, false, 3, 1));

		//		Composite composite_check = new Composite(group_complement, SWT.NONE);
		//		composite_check.setLayout(new RowLayout(SWT.HORIZONTAL));
		//		composite_check.setLayoutData(new GridData(SWT.LEFT, SWT.CENTER, false, false, 2, 1));

				btnCheckButton = new Button(group_complement, SWT.CHECK);
				btnCheckButton.setLayoutData(new GridData(SWT.LEFT, SWT.CENTER, true, false, 2, 1));
				btnCheckButton.addSelectionListener(new SelectionAdapter() {
					@Override
					public void widgetSelected(SelectionEvent e) {
						setComplementCheckBox(btnCheckButton.getSelection());
					}
				});
				btnCheckButton.setText("電話番号やアドレスを補完する");

				Label label_usertel = new Label(group_complement, SWT.NONE);
				label_usertel.setSize(84, 15);
				label_usertel.setText("ユーザ電話番号");

		text_usertel = new Text(group_complement, SWT.BORDER);
		text_usertel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		text_usertel.setSize(234, 21);

				Label label_useraddr = new Label(group_complement, SWT.NONE);
				label_useraddr.setSize(84, 15);
				label_useraddr.setText("ユーザアドレス");

		text_useraddr = new Text(group_complement, SWT.BORDER);
		text_useraddr.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, false, 1, 1));
		text_useraddr.setSize(246, 21);

		Composite composite_exec = new Composite(composite_main, SWT.NONE);
		composite_exec.setLayoutData(BorderLayout.SOUTH);
		composite_exec.setLayout(new GridLayout(3, false));

		Label label_exec = new Label(composite_exec, SWT.NONE);
		label_exec.setText("進捗");

		progressBar = new ProgressBar(composite_exec, SWT.NONE);
		progressBar.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
		progressBar.setMaximum(100);

		button_exec = new Button(composite_exec, SWT.NONE);
		button_exec.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				if (th_exec == null) {
					startExec();
				} else {
					// 停止
					button_exec.setEnabled(false);
					interrupted.set(true);
				}
			}
		});
		GridData gd_button_exec = new GridData(SWT.RIGHT, SWT.CENTER, false, false, 1, 1);
		gd_button_exec.widthHint = 70;
		button_exec.setLayoutData(gd_button_exec);
		button_exec.setText(exec_str);

		text_log = new Text(mainShell, SWT.MULTI | SWT.READ_ONLY | SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL);
		text_log.setLayoutData(BorderLayout.CENTER);

	}

	/**
	 * 実行ボタンにより実行可能かどうかを判断し、実行する
	 */
	private void startExec() {
		// 実行環境を設定
		String dir = text_dir.getText();
		targetDir = new File(dir);
		complement_mode = btnCheckButton.getSelection();
		usertel = text_usertel.getText();
		useraddr = text_useraddr.getText();
		// 設定を確認して保存
		if (usertel.contains("-")) {
			usertel = usertel.replaceAll("-", "");
			text_usertel.setText(usertel);
		}
		button_exec.setEnabled(false);
		progressPercent.set(0);
		progressBar.setState(SWT.NONE);
		saveSetting();
		text_log.setText("");
//		styledText.replaceTextRange(0, styledText.getCharCount(), "");
		if (dir.isEmpty() || complement_mode && (usertel.isEmpty() || useraddr.isEmpty())) {
			println("設定が不完全なため実行できません。", null);
		} else {
			th_exec = new Thread(myself);
			button_exec.setText(stop_str);
			th_exec.start();
		}
		button_exec.setEnabled(true);
	}

	/**
	 * 補完チェックボックスに連動して編集領域を変更する
	 * @param b
	 */
	private void setComplementCheckBox(boolean b) {
		btnCheckButton.setSelection(b);
		text_usertel.setEnabled(b);
		text_useraddr.setEnabled(b);
	}

	/**
	 * 設定をロードする
	 * @throws IOException
	 */
	private void loadSetting() {
		// プロパティファイルの存在を確認
		if (!SETTING_FILE.exists()) {
			text_dir.setText(Parameter.DIR.defvalue);
			setComplementCheckBox(Boolean.parseBoolean(Parameter.COMPLEMENT.defvalue));
			text_usertel.setText(Parameter.USERTEL.defvalue);
			text_useraddr.setText(Parameter.USERADDR.defvalue);
			sentbox = Parameter.SENTBOX.defvalue;
			saveSetting();
			text_log.append("設定ファイルが存在しないため作成しました。\n");
			return;
		}
		// 読み込み
		InputStream is = null;
		try {
			is = new FileInputStream(SETTING_FILE);
			Properties prop = new Properties();
			prop.loadFromXML(is);
			is.close();
			is = null;
			text_dir.setText(prop.getProperty(Parameter.DIR.propkey, Parameter.DIR.defvalue));
			setComplementCheckBox(Boolean.parseBoolean(prop.getProperty(Parameter.COMPLEMENT.propkey, Parameter.COMPLEMENT.defvalue)));
			text_usertel.setText(prop.getProperty(Parameter.USERTEL.propkey, Parameter.USERTEL.defvalue));
			text_useraddr.setText(prop.getProperty(Parameter.USERADDR.propkey, Parameter.USERADDR.defvalue));
			sentbox = prop.getProperty(Parameter.SENTBOX.propkey, Parameter.SENTBOX.defvalue);
		} catch (IOException e) {
			println(e, null);
			if (is != null) {
				try {
					is.close();
				} catch (IOException ex) {
					println(ex, null);
				}
			}
		}
	}

	/**
	 * 設定を保存する
	 * @throws IOException
	 */
	private void saveSetting() {
		OutputStream os = null;
		try {
			Properties prop = new Properties();
			prop.setProperty(Parameter.DIR.propkey, text_dir.getText());
			prop.setProperty(Parameter.COMPLEMENT.propkey, String.valueOf(btnCheckButton.getSelection()));
			prop.setProperty(Parameter.USERTEL.propkey, text_usertel.getText());
			prop.setProperty(Parameter.USERADDR.propkey, text_useraddr.getText());
			prop.setProperty(Parameter.SENTBOX.propkey, sentbox);
			os = new FileOutputStream(SETTING_FILE);
			prop.storeToXML(os, PROGRAM, "UTF-8");
			os.close();
			os = null;
		} catch (IOException e) {
			println(e, null);
			if (os != null) {
				try {
					os.close();
				} catch (IOException ex) {
					println(ex, null);
				}
			}
		}
	}


	/**
	 * ファイルを検知して変換
	 */
	@Override
	public void run() {
		interrupted.set(false);
		PrintWriter log = null;
		try {
			// ログファイル準備
			log = new PrintWriter(new File(targetDir, LOG_FILENAME));
			println(PROGRAM + " log file", log);
			println("実行日時: " + new Date(), log);
			println("変換元フォルダ: " + targetDir, log);
			println("補完モード: " + complement_mode, log);
			if (complement_mode) {
				println("補完ユーザ電話番号: " + usertel, log);
				println("補完ユーザアドレス: " + useraddr, log);
			}
			File[] files = targetDir.listFiles(vmgFileFilter);
			println(files.length + " 個のファイルを変換します。", log);

			// 変換処理
			for (int idx_file = 0 ; idx_file < files.length ; idx_file++) {
				File file = files[idx_file];
				println("変換中: " + file, log);
				String filename = file.getName();
				String outfile_prefix = filename.substring(0, filename.length() - 4); // 拡張子削除
				// ファイル内容読み込み
				List<VmgTuple> tuples = splitVmg(file);
				println(file.getName() + " に " + tuples.size() + " 通のメールを発見しました。", log);
				// メール解析
				int tuplesSize = tuples.size();
				for (int idx_tuple = 0 ; idx_tuple < tuplesSize ; idx_tuple++) {
					VmgTuple tuple = tuples.get(idx_tuple);
					if (interrupted.get()) {
						println(stop_str + "ボタンにより中断されました。", log);
						return;
					}
					File outfile = new File(targetDir, outfile_prefix + (tuplesSize == 1 ? ".eml" : "-" + String.format("%04d", idx_tuple + 1) + ".eml"));
					PrintWriter out = new PrintWriter(outfile, CHARSET); // SJISで出力
					String logline = convertToEml(tuple, out);
					out.close();
					if (logline != null) {
						println(outfile.getName() + ": " + logline, log);
					}
					// 進捗更新
					double div_file = 100 / files.length;
					double div_tuple = div_file / tuplesSize;
					progressPercent.set((int)(div_file * idx_file + div_tuple * (idx_tuple + 1)));
				}
			}
			progressPercent.set(100);
			println("変換完了", log);
		} catch (Exception e) {
			println(e, log);
		} finally {
			if (log != null) {
				log.close();
			}
			if (display != null) {
				display.asyncExec(run_button_exec);
			}
		}
	}

	/**
	 * vmgファイルを各tupleに分割
	 * @return 分割されたtuple
	 */
	private List<VmgTuple> splitVmg(File file) throws IOException {
		// ファイル内容読み込み
		BufferedReader br = null;
		try {
			br = new BufferedReader(new InputStreamReader(new FileInputStream(file), CHARSET));
			String line = null;
			List<VmgTuple> tuples = new ArrayList<VmgTuple>();
			VmgTuple tuple = null;
			while ((line = br.readLine()) != null) {
				if (tuple == null && line.equals("BEGIN:VMSG")) {
					tuple = new VmgTuple();
					tuple.meta = new ArrayList<String>();
					continue;
				}
				if (tuple == null) {
					continue;
				}
				if (tuple.meta != null && line.equals("BEGIN:VBODY")) {
					tuple.mailheader = new ArrayList<String>();
					continue;
				}
				if (tuple.mailheader != null && tuple.mailbody == null && line.equals("")) {
					tuple.finish_mailheader = true;
					tuple.mailbody = new ArrayList<String>();
					continue;
				}
				if (tuple.mailbody != null && line.equals("END:VBODY")) {
					tuple.finish_mailbody = true;
					continue;
				}
				if (tuple.meta != null && line.equals("END:VMSG")) {
					tuple.finish_meta = true;
					tuples.add(tuple);
//					System.out.println("[META]" + tuple.meta);
//					System.out.println("[Header]" + tuple.mailheader);
//					System.out.println("[Body]" + tuple.mailbody + "\n");
					tuple = null;
					continue;
				}
				if (tuple.mailbody != null && !tuple.finish_mailbody) {
					tuple.mailbody.add(line);
				} else if (tuple.mailheader != null && !tuple.finish_mailheader) {
					tuple.mailheader.add(line);
				} else if (tuple.meta != null && !tuple.finish_meta) {
					tuple.meta.add(line);
				} else {
					System.err.println("drop line: " + line);
				}
			}
			br.close();
			br = null;
			return tuples;
		} finally {
			if (br != null) {
				br.close();
			}
		}
	}

	/**
	 * vmgからemlに変換し、outに出力。ログに必要な情報はtupleに保存
	 * @param tuple
	 * @param out
	 * @return 変換結果ログ
	 */
	private String convertToEml(VmgTuple tuple, PrintWriter out) {
		List<String> newMeta = new LinkedList<String>();
		List<String> newHeader = new LinkedList<String>();
		List<String> newFrom = new ArrayList<String>();
		List<String> newTo = new ArrayList<String>();
		List<String> newCc = new ArrayList<String>();
		List<String> newBcc = new ArrayList<String>();
		boolean isSMS = false;
		String meta_type = null;
		String meta_box = null;
		String meta_tel = null;
		String header_subject = null;
		String logline = null;
		// メタ情報チェック
		for (String metaline : tuple.meta) {
			String[] kv = splitKV(metaline);
			if (kv[0].equals("X-IRMC-TYPE")) {
				meta_type = kv[1];
			}
			else if (kv[0].equals("X-IRMC-BOX")) {
				meta_box = kv[1];
			}
			else if (kv[0].equals("TEL") && !kv[1].equals("")) {
				meta_tel = kv[1];
			}
			newMeta.add("VMG-" + metaline);
		}
		// メールヘッダチェック
		for (String headerline : tuple.mailheader) {
			String[] kv = splitKV(headerline);
			if (kv[0].equals("Subject")) {
				header_subject = kv[1];
			}
			if (kv[0].equals("From")) {
				newFrom.add(kv[1]);
			}
			else if (kv[0].equals("To")) {
				newTo.add(kv[1]);
			}
			else if (kv[0].equals("Cc")) {
				newCc.add(kv[1]);
			}
			else if (kv[0].equals("Bcc")) {
				newBcc.add(kv[1]);
			}
			else {
				newHeader.add(headerline);
			}
		}

		// 補完
		if (complement_mode) {
			String complement_from = null;
			String complement_to = null;
			String complement_etc = null;
			// SMSの処理
			if (meta_type.equals("SMS") || newFrom.size() == 0 && meta_tel != null || newFrom.contains("iMS")) {
				isSMS = true;
				String desttel = null;
				if (meta_tel != null) {
					desttel = meta_tel;
					complement_from = "TELフィールドより補完";
				}
				if (desttel == null && header_subject != null) {
					Matcher matcher = regex_tel_from_subject.matcher(header_subject);
					if (matcher.matches() && matcher.groupCount() > 0) {
						desttel = matcher.group(1);
						complement_from = "Subjectフィールドより補完";
					}
				}
				if (desttel == null) {
					desttel = "";
				}
				newFrom.clear();
				newTo.clear();
				if (meta_box != null && meta_box.equals(sentbox)) {
					newFrom.add(usertel);
					newTo.add(desttel);
				} else {
					newFrom.add(desttel);
					newTo.add(usertel);
				}
				complement_to = complement_from;
				replaceField(newHeader, "Subject:", "Subject: (SMS:" + meta_box + ")", true);
				// キャラクタコード変換
				replaceField(newHeader, "CHARSET=SHIFT_JIS", "Content-Type: text/plain; charset=Shift_JIS", false);
				if (newHeader.contains("ENCODING=QUOTED-PRINTABLE")) {
					String oldBody = join(tuple.mailbody, "\n");
					oldBody = oldBody.replaceAll("=\n", ""); // =で終わる行は改行にしない
					try {
						tuple.mailbody.clear();
						tuple.mailbody.add(qpCodec.decode(oldBody));
						replaceField(newHeader, "ENCODING=QUOTED-PRINTABLE", "Content-Transfer-Encoding: 8bit", false);
						complement_etc = "本文の文字コード変更";
					} catch (DecoderException e) {
						// エラーがあったら古いエンコードをそのまま使う
						e.printStackTrace();
						complement_etc = "本文の文字コード変更に失敗";
					}
				}
			} else {
				// SMSじゃなかった場合
				// Fromヘッダ置換
				if (!isSMS && meta_box.equals(sentbox) && !newFrom.contains(useraddr)) {
					// Fromにユーザアドレスが無い場合、空のFromを消して置換
					if (newFrom.contains("")) {
						newFrom.remove("");
						complement_from = "空のFromを補完";
					} else {
						complement_from = "Fromに追加補完";
					}
					newFrom.add(useraddr);
				} else {
					complement_from = "補完せず";
				}
				// Toヘッダ置換
				if (!isSMS && !meta_box.equals(sentbox) && !newTo.contains(useraddr) && !newCc.contains(useraddr)) {
					// ToにもCcにもユーザアドレスが無い場合、空のToを消して置換
					if (newTo.contains("")) {
						newTo.remove("");
						complement_to = "空のToを補完";
					} else {
						complement_to = "Toに追加補完";
					}
					newTo.add(useraddr);
				} else {
					complement_to = "補完せず";
				}
			}
			logline = "[Box] " + meta_box + ", [isSMS] " + isSMS + ", [From] " + complement_from + ",[To] " + complement_to;
			if (complement_etc != null) {
				logline += ", [etc] " + complement_etc;
			}
			// meta情報出力
			for (String line : newMeta) {
				out.println(line);
			}
			// From/To/Cc/Bcc情報出力
			if (!newFrom.isEmpty()) {
				out.println("From: " + join(newFrom, ", "));
			}
			if (!newTo.isEmpty()) {
				out.println("To: " + join(newTo, ", "));
			}
			if (!newCc.isEmpty()) {
				out.println("Cc: " + join(newCc, ", "));
			}
			if (!newBcc.isEmpty()) {
				out.println("Bcc: " + join(newBcc, ", "));
			}
			// mailheader情報出力
			for (String line : newHeader) {
				out.println(line);
			}
		} else {
			// 補完しない場合
			// meta情報出力
			for (String line : tuple.meta) {
				out.println(line);
			}
			// mailheader情報出力
			for (String line : tuple.mailheader) {
				out.println(line);
			}
		}
		out.println();
		// mailbody情報出力
		for (String line : tuple.mailbody) {
			out.println(line);
		}
		return logline;
	}

	/**
	 * KeyとValueを分割する
	 * @param line
	 * @return
	 */
	private String[] splitKV(String line) {
		String[] ret = line.split(":", 2);
		ret[0] = ret[0].trim();
		if (ret.length == 2) {
			ret[1] = ret[1].trim();
		}
		return ret;
	}

	/**
	 * ヘッダ情報を置換する
	 * @param list
	 * @param oldprefix
	 * @param newline
	 * @return
	 */
	int replaceField(List<String> list, String oldprefix, String newline, boolean addIfNew) {
		int replacedNum = 0;
		for (int i=0 ; i < list.size(); i++) {
			if (list.get(i).startsWith(oldprefix)) {
				list.remove(i);
				if (newline != null) {
					list.add(i, newline);
				}
				replacedNum++;
			}
		}
		if (replacedNum == 0 && newline != null && addIfNew) {
			list.add(newline);
			replacedNum++;
		}
		return replacedNum;
	}

	/**
	 * 文字配列の結合
	 * @param s
	 * @param delimiter
	 * @return
	 */
	private String join(Collection<String> s, String delimiter) {
		StringBuffer buffer = new StringBuffer();
		Iterator<String> iter = s.iterator();
		while (iter.hasNext()) {
			buffer.append(iter.next());
			if (iter.hasNext()) {
				buffer.append(delimiter);
			}
		}
		return buffer.toString();
	}

	/**
	 * エラーを画面とログに表示
	 * @param e
	 * @param pw
	 */
	private void println(Exception e, PrintWriter pw) {
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		PrintWriter err_pw = new PrintWriter(new OutputStreamWriter(bos));
		e.printStackTrace(err_pw);
		String err = bos.toString();
		println(err, pw);
	}

	/**
	 * 画面とログに表示
	 * @param s
	 * @param pw
	 */
	private void println(String s, PrintWriter pw) {
		logPrinter.print(s + "\n");
		if (pw != null) {
			pw.println(s);
		}
	}

	/**
	 * メールデータセット
	 * @author ochi
	 *
	 */
	class VmgTuple {
		List<String> meta;
		List<String> mailheader;
		List<String> mailbody;
		boolean finish_meta;
		boolean finish_mailheader;
		boolean finish_mailbody;
	}

	/**
	 * 描画スレッドからログを表示するタスク
	 * @author ochi
	 */
	class LogPrinter implements Runnable {
		private StringBuilder sb = new StringBuilder();
		private int previousProgress = 0;

		public synchronized void print(String s) {
			if (display == null) {
				System.out.print(s);
				return;
			}
			if (sb.length() == 0 && s != null && s.length() > 0) {
				sb.append(s);
				display.asyncExec(this);
			} else {
				// 空じゃないということはasyncExec実行待ちということで、再度asyncExecしない
				sb.append(s);
			}
		}

		@Override
		public void run() {
			String drain = null;
			synchronized(this) {
				drain = sb.toString();
				sb.setLength(0);
			}
			int progress = progressPercent.get();
			if (progress != previousProgress) {
//				if (interrupted.get()) {
//					progressBar.setState(SWT.ERROR);
//					progressBar.redraw();
//				}
				progressBar.setSelection(progress);
				previousProgress = progress;
			}
			text_log.append(drain);
//			text_log.setTopIndex(text_log.getLineCount());
		}

	}

}
